import numpy as np
import pandas as pd

import bofire.outlier_detection.api as mapper
from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.features.api import (
    ContinuousInput,
    ContinuousOutput,
)
from bofire.data_models.kernels.api import RBFKernel, ScaleKernel
from bofire.data_models.outlier_detection.api import IterativeTrimming
from bofire.data_models.surrogates.api import ScalerEnum, SingleTaskGPSurrogate


def neal_func(x):
    return 0.3 + 0.4 * x + 0.5 * np.sin(2.7 * x) + 1.1 / (1 + x**2)


np.random.seed(5)


def test_IterativeTrimming():
    # generate mock data
    n = 200
    noise = 0.2
    n_outlier = 60
    noise_outlier = 1

    x_ob = np.random.rand(n) * 6 - 3
    y_ob = neal_func(x_ob) + np.random.randn(n) * noise
    y_ob[:n_outlier] = (
        neal_func(x_ob[:n_outlier]) + np.random.randn(n_outlier) * noise_outlier
    )

    experiments = pd.DataFrame()
    experiments["x_1"] = x_ob
    experiments["y"] = y_ob
    experiments["valid_y"] = 1
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=experiments.keys()[0],
                bounds=(-3, 3),
            )
            for i in range(1)
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key=experiments.keys()[1])])
    kernel = ScaleKernel(base_kernel=RBFKernel(ard=True))
    scaler = ScalerEnum.NORMALIZE
    ITGP_model = IterativeTrimming(
        base_gp=SingleTaskGPSurrogate(
            inputs=inputs, outputs=outputs, kernel=kernel, scaler=scaler
        )
    )
    ITGP = mapper.map(ITGP_model)
    assert isinstance(ITGP.base_gp, SingleTaskGPSurrogate)
    assert isinstance(ITGP, mapper.IterativeTrimming)
    # detect
    trimmed, outliers = ITGP.detect(experiments=experiments)
    assert isinstance(trimmed, pd.DataFrame)
    assert isinstance(outliers, pd.DataFrame)
    assert len(experiments) == len(trimmed) + len(outliers)
    assert len(outliers) >= n_outlier / 2



import importlib
import random
import warnings

import bofire.data_models.molfeatures.api as molfeatures
from tests.bofire.data_models.specs.specs import Specs

try:
    from rdkit.Chem import Descriptors

    fragments_list = [fragment[0] for fragment in Descriptors.descList[124:]]
except ImportError:
    warnings.warn(
        "rdkit not installed, BoFire's cheminformatics utilities cannot be used."
    )

try:
    from mordred import Calculator, descriptors

    calc = Calculator(descriptors, ignore_3D=False)
    mordred_descriptors = [str(d) for d in calc.descriptors]
except ImportError:
    warnings.warn(
        "mordred not installed. Mordred molecular descriptors cannot be used."
    )

RDKIT_AVAILABLE = importlib.util.find_spec("rdkit") is not None
MORDRED_AVAILABLE = importlib.util.find_spec("mordred") is not None

specs = Specs([])

specs.add_valid(
    molfeatures.Fingerprints,
    lambda: {
        "bond_radius": random.randrange(1, 6),
        "n_bits": random.randrange(32, 2048),
    },
)

if RDKIT_AVAILABLE:
    specs.add_valid(
        molfeatures.Fragments,
        lambda: {
            "fragments": random.sample(
                fragments_list, k=random.randrange(1, len(fragments_list))
            )
        },
    )
    specs.add_valid(
        molfeatures.FingerprintsFragments,
        lambda: {
            "bond_radius": random.randrange(1, 6),
            "n_bits": random.randrange(32, 2048),
            "fragments": random.sample(
                fragments_list, k=random.randrange(1, len(fragments_list))
            ),
        },
    )

    if MORDRED_AVAILABLE:
        specs.add_valid(
            molfeatures.MordredDescriptors,
            lambda: {
                "descriptors": random.sample(
                    mordred_descriptors, k=random.randrange(1, 10)
                )
            },
        )



from typing import Literal, Sequence

from bofire.data_models.surrogates.botorch import BotorchSurrogate
from bofire.data_models.surrogates.scaler import ScalerEnum
from bofire.data_models.surrogates.trainable import TrainableSurrogate


class MLPEnsemble(BotorchSurrogate, TrainableSurrogate):
    type: Literal["MLPEnsemble"] = "MLPEnsemble"
    n_estimators: int
    hidden_layer_sizes: Sequence = (100,)
    activation: Literal["relu", "logistic", "tanh"] = "relu"
    dropout: float = 0.0
    batch_size: int = 10
    n_epochs: int = 200
    lr: float = 1e-4
    weight_decay: float = 0.0
    subsample_fraction: float = 1.0
    shuffle: bool = True
    scaler: ScalerEnum = ScalerEnum.NORMALIZE



import numpy as np

import bofire.surrogates.api as surrogates
from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.features.api import ContinuousInput, ContinuousOutput
from bofire.data_models.kernels.api import LinearKernel
from bofire.data_models.surrogates.api import LinearSurrogate


def test_LinearSurrogate():
    N_EXPERIMENTS = 10

    inputs = Inputs(
        features=[
            ContinuousInput(key="a", bounds=(0, 40)),
            ContinuousInput(key="b", bounds=(20, 60)),
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="c")])

    experiments = inputs.sample(N_EXPERIMENTS)
    experiments["c"] = (
        experiments["a"] * 2.2
        + experiments["b"] * -0.05
        + experiments["b"]
        + np.random.normal(loc=0, scale=5, size=N_EXPERIMENTS)
    )
    experiments["valid_c"] = 1

    surrogate_data = LinearSurrogate(inputs=inputs, outputs=outputs)
    surrogate = surrogates.map(surrogate_data)

    assert isinstance(surrogate, surrogates.SingleTaskGPSurrogate)
    assert isinstance(surrogate.kernel, LinearKernel)



from typing import Literal

from pydantic import Field
from typing_extensions import Annotated

from bofire.data_models.base import BaseModel
from bofire.data_models.surrogates.api import SingleTaskGPSurrogate


class OutlierDetection(BaseModel):
    type: str


class IterativeTrimming(OutlierDetection):
    """Remove outliers using Robust GP

    Paper: Robust Gaussian Process Regression Based on Iterative Trimming.
    https://arxiv.org/pdf/2011.11057.pdf

    Parameters
    ----------

    alpha1, alpha2: float in (0, 1)
        Trimming and reweighting parameters respectively.
    nsh, ncc, nrw: int (>=1)
        Number of shrinking, concentrating, and reweighting iterations respectively.
    base_gp: SingleTaskGPSurrogate = Gaussian process model for outlier detection.
    """

    type: Literal["IterativeTrimming"] = "IterativeTrimming"
    alpha1: Annotated[float, Field(gt=0.0, lt=1.0)] = 0.5
    alpha2: Annotated[float, Field(gt=0.0, lt=1.0)] = 0.975
    nsh: Annotated[int, Field(ge=1)] = 2
    ncc: Annotated[int, Field(ge=1)] = 2
    nrw: Annotated[int, Field(ge=1)] = 1
    base_gp: SingleTaskGPSurrogate



import warnings
from typing import Optional, Tuple

import pandas as pd

import bofire.strategies.api as strategies
import bofire.surrogates.api as surrogates
from bofire.benchmarks.benchmark import Benchmark, run
from bofire.data_models.domain.api import Domain
from bofire.data_models.enum import RegressionMetricsEnum
from bofire.data_models.objectives.api import MinimizeObjective
from bofire.data_models.strategies.api import (
    FactorialStrategy,
    RandomStrategy,
    SoboStrategy,
)
from bofire.data_models.surrogates.api import AnyTrainableSurrogate


class Hyperopt(Benchmark):
    def __init__(
        self,
        surrogate_data: AnyTrainableSurrogate,
        training_data: pd.DataFrame,
        folds: int,
        random_state: Optional[int] = None,
    ) -> None:
        super().__init__()
        if surrogate_data.hyperconfig is None:
            raise ValueError("No hyperoptimization configuration found.")
        self.surrogate_data = surrogate_data
        self.training_data = training_data
        self.folds = folds
        self.results = None
        self.random_state = random_state

    @property
    def domain(self) -> Domain:
        return self.surrogate_data.hyperconfig.domain  # type: ignore

    @property
    def target_metric(self):
        return self.surrogate_data.hyperconfig.target_metric  # type: ignore

    def _f(self, candidates: pd.DataFrame) -> pd.DataFrame:
        for i, candidate in candidates.iterrows():
            self.surrogate_data.update_hyperparameters(candidate)
            surrogate = surrogates.map(self.surrogate_data)
            _, cv_test, _ = surrogate.cross_validate(  # type: ignore
                self.training_data, folds=self.folds, random_state=self.random_state
            )
            if i == 0:
                results = cv_test.get_metrics(combine_folds=True)
            else:
                results = pd.concat(
                    [results, cv_test.get_metrics(combine_folds=True)],  # type: ignore
                    ignore_index=True,
                    axis=0,
                )
        results[f"valid_{self.target_metric.value}"] = 1  # type: ignore
        return results  # type: ignore


def hyperoptimize(
    surrogate_data: AnyTrainableSurrogate,
    training_data: pd.DataFrame,
    folds: int,
    random_state: Optional[int] = None,
) -> Tuple[AnyTrainableSurrogate, pd.DataFrame]:
    if surrogate_data.hyperconfig is None:
        warnings.warn(
            "No hyperopt is possible as no hyperopt config is available. Returning initial config."
        )
        return surrogate_data, pd.DataFrame({e.name: [] for e in RegressionMetricsEnum})

    def best(domain: Domain, experiments: pd.DataFrame) -> float:
        return (
            experiments[domain.outputs[0].key].min()
            if isinstance(domain.outputs[0].objective, MinimizeObjective)
            else experiments[domain.outputs[0].key].max()
        )

    def sample(domain):
        datamodel = RandomStrategy(domain=domain)
        sampler = strategies.map(data_model=datamodel)
        sampled = sampler.ask(len(domain.inputs) + 1)
        return sampled

    benchmark = Hyperopt(
        surrogate_data=surrogate_data,
        training_data=training_data,
        folds=folds,
        random_state=random_state,
    )

    if surrogate_data.hyperconfig.hyperstrategy == "FactorialStrategy":  # type: ignore
        strategy = strategies.map(FactorialStrategy(domain=benchmark.domain))
        experiments = benchmark.f(
            strategy.ask(candidate_count=None), return_complete=True
        )
    else:
        experiments = run(
            benchmark=benchmark,
            strategy_factory=RandomStrategy
            if surrogate_data.hyperconfig.hyperstrategy == "RandomStrategy"  # type: ignore
            else SoboStrategy,  # type: ignore
            metric=best,
            n_runs=1,
            n_iterations=surrogate_data.hyperconfig.n_iterations  # type: ignore
            - len(benchmark.domain.inputs)
            - 1,
            initial_sampler=sample,
            n_procs=1,
        )[0][0]

    # analyze the results and get the best
    experiments = experiments.sort_values(
        by=benchmark.target_metric.name,
        ascending=True
        if isinstance(benchmark.domain.outputs[0].objective, MinimizeObjective)
        else False,
    )

    surrogate_data.update_hyperparameters(experiments.iloc[0])

    return (
        surrogate_data,
        experiments[
            surrogate_data.hyperconfig.domain.inputs.get_keys()
            + [e.name for e in RegressionMetricsEnum]
        ],
    )



import random

import bofire.data_models.acquisition_functions.api as acquisition_functions
from tests.bofire.data_models.specs.specs import Specs

specs = Specs([])

specs.add_valid(
    acquisition_functions.qEI,
    lambda: {},
)

specs.add_valid(
    acquisition_functions.qNEI,
    lambda: {},
)

specs.add_valid(
    acquisition_functions.qPI,
    lambda: {
        "tau": random.random(),
    },
)

specs.add_valid(
    acquisition_functions.qSR,
    lambda: {},
)

specs.add_valid(
    acquisition_functions.qUCB,
    lambda: {
        "beta": random.random(),
    },
)



from typing import Literal

from pydantic import Field

from bofire.data_models.kernels.api import LinearKernel
from bofire.data_models.priors.api import (
    BOTORCH_NOISE_PRIOR,
    BOTORCH_SCALE_PRIOR,
    AnyPrior,
)

# from bofire.data_models.strategies.api import FactorialStrategy
from bofire.data_models.surrogates.botorch import BotorchSurrogate
from bofire.data_models.surrogates.scaler import ScalerEnum
from bofire.data_models.surrogates.trainable import TrainableSurrogate


class LinearSurrogate(BotorchSurrogate, TrainableSurrogate):
    type: Literal["LinearSurrogate"] = "LinearSurrogate"

    kernel: LinearKernel = Field(
        default_factory=lambda: LinearKernel(variance_prior=BOTORCH_SCALE_PRIOR())
    )
    noise_prior: AnyPrior = Field(default_factory=lambda: BOTORCH_NOISE_PRIOR())
    scaler: ScalerEnum = ScalerEnum.NORMALIZE



from typing import Union

from bofire.data_models.outlier_detection.outlier_detection import (
    IterativeTrimming,
    OutlierDetection,
)

AnyOutlierDetection = Union[OutlierDetection, IterativeTrimming]
